#include <stdio.h>
#include <assert.h>
#include <sys/sysctl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>

#include <OpenCL/OpenCL.h>

#include "Image.h"
#include "math.h"
#include <cstdlib>
#include <fstream>
#include <iostream>


#pragma mark -
#pragma mark Utilities

int *inputImage; //2D array of int
int inputWidth;
int inputHeight;
int inputMaskSize;
int *inputMask; //2D array of int
gdImagePtr gdOut;

double getClock()
{
    struct timeval tp;
    struct timezone tzp;

    gettimeofday(&tp, &tzp);

    return (double)tp.tv_sec+((double)tp.tv_usec/1000000.0);
}


char * load_program_source(const char *filename)
{ 
	
	struct stat statbuf;
	FILE *fh; 
	char *source; 
	
	fh = fopen(filename, "r");
	if (fh == 0)
		return 0; 
	
	stat(filename, &statbuf);
	source = (char *) malloc(statbuf.st_size + 1);
	fread(source, statbuf.st_size, 1, fh);
	source[statbuf.st_size] = '\0'; 
	
	return source; 
} 

#pragma mark -
#pragma mark Main OpenCL Routine
int runCL(int *inputImage, int *inputMask, int *results)
{
	cl_program program[1];
	cl_kernel kernel[2];
	
	cl_command_queue cmd_queue;
	cl_context   context;
	
	cl_device_id cpu = NULL, device = NULL;

	cl_int err = 0;

	size_t buffer_size;

	cl_mem inputImage_mem, inputMask_mem, answer_mem;
	
#pragma mark Device Information
	{
		// Find the CPU CL device, as a fallback
		err = clGetDeviceIDs(NULL, CL_DEVICE_TYPE_CPU, 1, &cpu, NULL);
		assert(err == CL_SUCCESS);
		
		// Find the GPU CL device, this is what we really want
		// If there is no GPU device is CL capable, fall back to CPU
		err = clGetDeviceIDs(NULL, CL_DEVICE_TYPE_GPU, 1, &device, NULL);
		if (err != CL_SUCCESS) device = cpu;
		assert(device);
	
		// Get some information about the returned device
		size_t returned_size = 0;
		cl_char vendor_name[1024] = {0};
		cl_char device_name[1024] = {0};
		err = clGetDeviceInfo(device, CL_DEVICE_VENDOR, sizeof(vendor_name), 
							  vendor_name, &returned_size);
		err |= clGetDeviceInfo(device, CL_DEVICE_NAME, sizeof(device_name), 
							  device_name, &returned_size);

		assert(err == CL_SUCCESS);
		printf("Connecting to %s %s...\n", vendor_name, device_name);
	}
	
#pragma mark Context and Command Queue
	{
		// Now create a context to perform our calculation with the 
		// specified device 
		context = clCreateContext(0, 1, &device, NULL, NULL, &err);
		assert(err == CL_SUCCESS);
		
		// And also a command queue for the context
		cmd_queue = clCreateCommandQueue(context, device, 0, NULL);
	}
	
#pragma mark Program and Kernel Creation
	{
		// Load the program source from disk
		// The kernel/program is the project directory and in Xcode the executable
		// is set to launch from that directory hence we use a relative path
		const char * filename = "convolute.cl";
		char *program_source = load_program_source(filename);
		program[0] = clCreateProgramWithSource(context, 1, (const char**)&program_source,
											   NULL, &err);
		assert(err == CL_SUCCESS);
		
		err = clBuildProgram(program[0], 0, NULL, NULL, NULL, NULL);
		assert(err == CL_SUCCESS);
		
		// Now create the kernel "objects" that we want to use in the example file 
		kernel[0] = clCreateKernel(program[0], "convolute", &err);
	}
		
#pragma mark Memory Allocation
	{
		// Allocate memory on the device to hold our data and store the results into
		buffer_size = sizeof(int) * inputWidth * inputHeight;
		
		// This will contain the inputImage so it can be read only
		inputImage_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, &err);
		assert(err == CL_SUCCESS);

		err = clEnqueueWriteBuffer(cmd_queue, inputImage_mem, CL_TRUE, 0, buffer_size,
								   (void*) inputImage, 0, NULL, NULL);
		assert(err == CL_SUCCESS);
		
		// This will contain the inputMask. Read_Only
		buffer_size = sizeof(int) * inputMaskSize * inputMaskSize;
		inputMask_mem = clCreateBuffer(context, CL_MEM_READ_ONLY, buffer_size, NULL, NULL);
		err = clEnqueueWriteBuffer(cmd_queue, inputMask_mem, CL_TRUE, 0, buffer_size,
									(void*) inputMask, 0, NULL, NULL);
		assert(err == CL_SUCCESS);
		
		// Results array
		buffer_size = sizeof(int) * inputWidth * inputHeight;
		answer_mem	= clCreateBuffer(context, CL_MEM_READ_WRITE, buffer_size, NULL, NULL);
		
		// Blocking call to make sure the device is done allocating our 3-buffers.
		clFinish(cmd_queue);
	}
	
#pragma mark Kernel Arguments
	{
		// Now setup the arguments to our kernel
		err  = clSetKernelArg(kernel[0],  0, sizeof(cl_mem), &inputImage_mem);
		assert(err == CL_SUCCESS);
		err = clSetKernelArg(kernel[0],  1, sizeof(cl_mem), &inputMask_mem);
		assert(err == CL_SUCCESS);
		err = clSetKernelArg(kernel[0],  2, sizeof(cl_mem), &answer_mem);
		assert(err == CL_SUCCESS);
		err = clSetKernelArg(kernel[0],  3, sizeof(int), &inputWidth);
		assert(err == CL_SUCCESS);
		err = clSetKernelArg(kernel[0],  4, sizeof(int), &inputMaskSize);
		assert(err == CL_SUCCESS);
	}
	
#pragma mark Execution and Read
	{
		// Run the calculation by enqueuing it and forcing the 
		// command queue to complete the task
		size_t global_work_size[2];
		global_work_size[0] = inputWidth; // Number of computation to perform.
		global_work_size[1] = inputHeight; // Number of computation to perform.

		//Create the range of our work.. 2D problem! Let openCL decide the local_work_size (null)
		err = clEnqueueNDRangeKernel(cmd_queue, kernel[0], 2, NULL, 
									 global_work_size, NULL, 0, NULL, NULL);
		assert(err == CL_SUCCESS);

		clFinish(cmd_queue);
		
		// Once finished read back the results from the answer 
		// array into the results array
		err = clEnqueueReadBuffer(cmd_queue, answer_mem, CL_TRUE, 0, buffer_size, 
								  results, 0, NULL, NULL);
		if (err == CL_INVALID_COMMAND_QUEUE)
			printf("\n!1\n");
		if (err == CL_INVALID_CONTEXT)
			printf("\n!2\n");
		if (err == CL_INVALID_MEM_OBJECT)
			printf("\n!3\n");
		if (err == CL_INVALID_VALUE)
			printf("\n!4\n");
		if (err == CL_INVALID_EVENT_WAIT_LIST)
			printf("\n!5\n");
		if (err == CL_MEM_OBJECT_ALLOCATION_FAILURE)
			printf("\n!6\n");
		if (err == CL_OUT_OF_HOST_MEMORY)
			printf("\n!7\n");
		assert(err == CL_SUCCESS);
		clFinish(cmd_queue);
	}
	
#pragma mark Teardown
	{
		clReleaseMemObject(inputImage_mem);
		clReleaseMemObject(inputMask_mem);
		clReleaseMemObject(answer_mem);
		
		clReleaseCommandQueue(cmd_queue);
		clReleaseContext(context);
	}
	return CL_SUCCESS;
}


void readSourceImage()
{

		FILE *I_in;
        gdImagePtr gdIn = 0;
        gdOut = 0;

        I_in = fopen ("Input.gif", "rb");
        if (!I_in) {
              cout << "Error: Cannot open input file." << endl;
              exit(-1);
        }
        gdIn = gdImageCreateFromGif(I_in);
        fclose(I_in);
        if (!gdIn) {
              cout << "Error: Not a valid input gif file." << endl;
              exit(-1);
        }

        Image imageIn(gdIn);
        Image imageOut(imageIn);

		ifstream M_in("Mask.in", ios::in);
		int mx, my;
		M_in >> mx >> my;
		inputMaskSize = my;

		inputWidth = imageIn.dx();
		inputHeight = imageIn.dy();
		printf("Input image is %iX%i\n",inputWidth, inputHeight);


		inputMask = (int *) malloc(inputMaskSize*inputMaskSize*sizeof(int));

		//Fill in the mask
        for (int y = 0; y < inputMaskSize; y++)
               for (int x = 0; x < mx; x++)
					M_in >> inputMask[y*inputMaskSize+x];

		inputImage = (int *) malloc(inputWidth*inputHeight*sizeof(int));
		//Fill in the inputImage array
		for (int y= 0; y< inputHeight; y++)
			for (int x= 0; x< inputWidth; x++)
				inputImage[y*inputWidth+x] = imageIn.get(x,y);


	    gdOut = gdImageCreate(inputWidth, inputHeight);
	    if (!gdOut){
	          cout << "Error: Unable to create output gd file!" << endl;
	          exit(-1);
	    }

	    for (int i = 0; i < gdMaxColors; i++) {
	          gdImageColorAllocate(gdOut, gdImageRed(gdIn, i), gdImageGreen(gdIn, i),gdImageBlue(gdIn, i));
	    }

		printf("Setting pixel");
		gdImageSetPixel(gdOut, 1, 1, 1);
		printf("Pixel set");

		printf("Setting pixel");
		gdImageSetPixel(gdOut, 1, 1, 1);
		printf("Pixel set");

}


int main (int argc, const char * argv[]) {
    
	// This will actually read the image and mask into inputImage[][] and inputMask[][]
	readSourceImage();


	printf("!! going for an image %iX%i. Mask:%i\n", inputWidth, inputHeight, inputMaskSize);
	
	// Allocate some memory and a place for the results
	int *results = (int *) malloc(sizeof(int)*inputWidth*inputHeight);

	double start, end;
	start = getClock();

	// Do the OpenCL calculation
	runCL(inputImage, inputMask, results);

	end = getClock();

	printf("Done convoluting in %f seconds.\n",(end-start));
	printf("I survived and I'm done! Writing output.\n");

	//Time to record MY values...
    for (int y = 0; y < inputHeight; y++)
          for (int x = 0; x < inputWidth; x++)
                 gdImageSetPixel(gdOut, x, y, results[y*inputWidth+x]);



	FILE *I_out;
    I_out = fopen ("Output.gif", "wb");
    if (!I_out){
          cout << "Error: Unable to create output gif file." << endl;
          exit (-1);
    }
    gdImageGif(gdOut, I_out);
    fclose(I_out);

	gdImageDestroy(gdOut);

	
	// Free up memory
	free(inputImage);
	free(inputMask);
	free(results);
	
    return 0;
}









